name = 'qt_thread_updater'
version = '1.1.2'
description = 'Python Qt thread updater to update GUI items using a separate thread.'
url = 'https://github.com/justengel/qt_thread_updater'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'
